from typing import TYPE_CHECKING, Optional

from httpx._models import Cookies

if TYPE_CHECKING:
    from .user import Me

class Settings:
    '''`bestdori_api` 设置项类'''
    
    proxy: Optional[str] = None
    '''代理服务器'''
    cookies: Optional[Cookies] = None
    '''Cookies

    不建议直接设置，推荐通过 `Settings.set_cookies` 方法，
    传入一个 `bestdori.user.Me` 对象设置
    '''
    print: bool = False
    '''是否打印输出信息'''
    
    def set_cookies(self, me: 'Me') -> 'Settings':
        '''设置 Cookies'''
        self.cookies = me.cookies
        return self

settings = Settings()



'''`bestdori.gacha`

BanG Dream! 招募相关操作'''
from typing import Optional, Literal, Any

from .post import get_list
from ._settings import settings
from .utils.utils import API, ASSETS
from .utils.network import Api, Assets
from .exceptions import (
    ServerNotAvailableError,
    AssetsNotExistError,
    GachaNotExistError
)

# 获取总招募信息
def get_all(index: Literal[0, 5]=5) -> dict[str, dict[str, Any]]:
    '''获取总招募信息

    参数:
        index (Literal[0, 5], optional): 指定获取哪种 `all.json`
            `0`: 仅获取所有已有招募 ID `all.0.json`
            `5`: 获取所有已有招募的简洁信息 `all.5.json`

    返回:
        dict[str, dict[str, Any]]: 获取到的总招募信息
    '''
    return Api(API['gacha']['all'].format(index), settings.proxy).request('get').json()

# 招募类
class Gacha:
    '''招募类

    参数:
        id_ (int): 招募 ID
    '''
    # 初始化
    def __init__(self, id_: int) -> None:
        '''招募类

        参数:
            id_ (int): 招募 ID
        '''
        self.id: int = id_
        '''招募 ID'''
        self._info: dict[str, Any] = {}
        '''招募信息'''
        # 检测 ID 是否存在
        all_id = get_all(0)
        if not id_ in all_id.keys():
            raise GachaNotExistError(id_)
        return
    
    # 获取招募信息
    def get_info(self) -> dict[str, Any]:
        '''获取招募信息

        返回:
            dict[str, Any]: 招募详细信息
        '''
        if len(self._info) <= 0:
            # 如果没有招募信息存储
            response = Api(
                API['gacha']['info'].format(self.id), settings.proxy
            ).request('get')
            self._info = dict(response.json())
        return self._info
    
    # 获取招募评论
    def get_comment(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_ASC'
    ) -> dict[str, Any]:
        '''获取招募评论

        参数:
            limit (int, optional): 展示出的评论数，默认为 20
            offset (int, optional): 忽略前面的 `offset` 条评论，默认为 0
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 排序顺序，默认时间顺序

        返回:
            dict[str, Any]: 搜索结果
            ```python
            result: bool # 是否有响应
            count: int # 搜索到的评论总数
            posts: list[dict[str, Any]] # 列举出的评论
            ```
        '''
        return get_list(
            category_name='GACHA_COMMENT',
            category_id=str(self.id),
            order=order,
            limit=limit,
            offset=offset
        )
    
    # 获取招募标题
    @property
    def name(self) -> str:
        '''获取招募标题

        返回:
            str: 招募标题
        '''
        info = self.get_info()
        # 获取 eventName 数据
        if (gacha_name := info.get('gachaName', None)) is None:
            raise Exception('无法获取招募标题。')
        # 获取第一个非 None 招募标题
        try:
            return next(filter(lambda x: x is not None, gacha_name))
        except StopIteration:
            raise Exception('无法获取招募标题。')
    
    # 获取招募默认服务器
    @property
    def server(self) -> Literal['jp', 'en', 'tw', 'cn', 'kr']:
        '''获取招募默认服务器

        返回:
            Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]: 歌曲所在服务器
        '''
        info = self.get_info()
        # 获取 publishedAt 数据
        if (published_at := info.get('publishedAt', None)) is None:
            raise Exception('无法获取招募起始时间。')
        # 根据 publishedAt 数据判断服务器
        if published_at[0] is not None: return 'jp'
        elif published_at[1] is not None: return 'en'
        elif published_at[2] is not None: return 'tw'
        elif published_at[3] is not None: return 'cn'
        elif published_at[4] is not None: return 'kr'
        else:
            raise Exception('无法获取招募所在服务器。')
    
    # 获取招募缩略图图片
    def get_banner(self, server: Literal['jp', 'en', 'tw', 'cn', 'kr']) -> bytes:
        '''获取招募缩略图图片

        参数:
            server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): 指定服务器

        返回:
            bytes: 招募缩略图图片字节数据 `bytes`
        '''
        # 获取招募数据包名称
        info = self.get_info()
        if (banner_asset_bundle_name := info.get('bannerAssetBundleName', None)) is None:
            raise ValueError('无法获取招募数据包名称。')
        # 判断服务器
        if (start_at := info.get('startAt', None)) is None:
            raise ValueError('无法获取招募起始时间。')
        SERVERS = ['jp', 'en', 'tw', 'cn', 'kr']
        index = SERVERS.index(server)
        if start_at[index] is None:
            raise ServerNotAvailableError(f'招募 {self.name}', server)
        return Assets(
            ASSETS['homebanner']['get'].format(
                banner_asset_bundle_name=banner_asset_bundle_name
            ), server, settings.proxy
        ).get()
    
    # 获取招募 pickup 图像
    def get_pickups(self, server: Literal['jp', 'en', 'tw', 'cn', 'kr']) -> list[bytes]:
        '''获取招募 pickup 图像

        参数:
            server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): _description_

        返回:
            list[bytes]: pickup 图像字节数据 `bytes` 列表
        '''
        PICKUPS = ['pickup1', 'pickup2', 'pickup']
        # 遍历尝试获取
        pickup_list: list[bytes] = []
        for pickup in PICKUPS:
            try:
                pickup_list.append(
                    Assets(
                        ASSETS['gacha']['screen'].format(
                            id=self.id, asset_name=pickup
                        ), server, settings.proxy
                    ).get()
                )
            except:
                continue
        if len(pickup_list) <= 0:
            # 没有获取到任何 pickup 图像
            raise AssetsNotExistError('招募 pickup 图像')
        return pickup_list
    
    # 获取招募 logo 图像
    def get_logo(self, server: Literal['jp', 'en', 'tw', 'cn', 'kr']) -> bytes:
        '''获取招募 logo 图像

        参数:
            server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): _description_

        返回:
            bytes: logo 图像字节数据 `bytes`
        '''
        try:
            return Assets(
                ASSETS['gacha']['screen'].format(
                    id=self.id, asset_name='logo'
                ), server, settings.proxy
            ).get()
        except:
            raise AssetsNotExistError('招募 logo 图像')



'''`bestdori.eventarchives`

BanG Dream! 活动数据相关操作'''
from typing import Optional, Literal, Any

from .post import get_list
from .utils.utils import API
from .utils.network import Api
from ._settings import settings
from .exceptions import (
    EventNotExistError
)

# 获取总活动数据信息
def get_all(index: Literal[5]=5) -> dict[str, dict[str, Any]]:
    '''获取总活动信息

    参数:
        index (Literal[5], optional): 指定获取哪种 `all.json`
            `5`: 获取所有已有活动数据的简洁信息 `all.5.json`

    返回:
        dict[str, dict[str, Any]]: 获取到的总活动信息
    '''
    return Api(API['all']['archives'].format(index), settings.proxy).request('get').json()

# 活动数据类
class EventArchive():
    '''活动数据类

    参数:
        id_ (int): 活动 ID
    '''
    # 初始化
    def __init__(self, id_: int) -> None:
        '''活动数据类

        参数:
            id_ (int): 活动 ID
        '''
        self.id: int = id_
        '''活动 ID'''
        self._info: dict[str, Any] = {}
        '''活动信息'''
        # 检测 ID 是否存在
        all_id = get_all(5)
        if not str(id_) in all_id.keys():
            raise EventNotExistError(id_)
        return
    
    # 获取排名分数线
    def get_top(
        self,
        server: Literal[0, 1, 2, 3, 4],
        mid: Literal['0']='0',
        latest: Literal['1']='1'
    ) -> dict[str, list[dict[str, Any]]]:
        '''获取排名分数线

        参数:
            server (Literal[0, 1, 2, 3, 4]): 指定服务器
                `0`: 日服
                `1`: 英服
                `2`: 台服
                `3`: 国服
                `4`: 韩服
            mid (Literal[&#39;0&#39;], optional): 指定是否为中间分数线，默认为 `0`
            latest (Literal[&#39;1&#39;], optional): 指定是否为最终分数线，默认为 `1`

        返回:
            dict[str, list[dict[str, Any]]]: 排名分数线数据
        '''
        return Api(API['events']['top'], settings.proxy).request(
            'get', params={
                'server': server,
                'event': self.id,
                'mid': mid,
                'latest': latest
            }
        ).json()

    # 获取活动数据评论
    def get_comment(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_ASC'
    ) -> dict[str, Any]:
        '''获取动数据评论

        参数:
            limit (int, optional): 展示出的评论数，默认为 20
            offset (int, optional): 忽略前面的 `offset` 条评论，默认为 0
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 排序顺序，默认时间顺序

        返回:
            dict[str, Any]: 搜索结果
            ```python
            result: bool # 是否有响应
            count: int # 搜索到的评论总数
            posts: list[dict[str, Any]] # 列举出的评论
            ```
        '''
        return get_list(
            category_id=str(self.id),
            category_name='EVENTARCHIVE_COMMENT',
            limit=limit,
            offset=offset,
            order=order
        )



'''`bestdori.ayachan`

ayachan 的各种 API 调用整合'''
from pathlib import Path
from mimetypes import guess_type
from typing import Optional, Literal, Union, Any

from ..charts import Chart
from .utils import Api, API
from ._settings import settings

# 自定义谱面分析
def chart_analysis(map_: Chart, diff: Literal['0', '1', '2', '3', '4']) -> dict[str, Any]:
    '''自定义谱面分析

    参数:
        map_ (Chart): 谱面
        diff (Literal[&#39;0&#39;, &#39;1&#39;, &#39;2&#39;, &#39;3&#39;, &#39;4&#39;]): 难度

    返回:
        dict[str, Any]: 分析结果
    '''
    payload = {
        'options': {
            'diff': diff
        },
        'map': map_.to_list(),
        'map_format_in': 'BestdoriV2'
    }
    return Api(API['map-info'], settings.proxy).request('post', data=payload).json()

# Bestdori 谱面分析
def bestdori_chart_analysis(
    id_: int,
    diff: Literal['0', '1', '2', '3', '4']
) -> dict[str, Any]:
    '''Bestdori 谱面分析

    参数:
        id_ (int): 谱面 ID
        diff (Literal[&#39;0&#39;, &#39;1&#39;, &#39;2&#39;, &#39;3&#39;, &#39;4&#39;]): 难度

    返回:
        dict[str, Any]: 分析结果
    '''
    return Api(API['bestdori'].format(id=id_), settings.proxy).request(
        'get',
        params={
            'diff': diff
        }
    ).json()

# Sonolus 谱面测试
def sonolus_test(
    title: str,
    bgm: Union[str, Path],
    difficulty: int,
    hidden: bool,
    lifetime: Literal[3600, 10800, 21600, 43200, 86400],
    chart: Chart
) -> int:
    '''Sonolus 谱面测试

    参数:
        title (str): 谱面标题
        bgm (Union[str, Path]): 音乐文件
        difficulty (int): 谱面难度
        hidden (bool): 谱面隐藏
        lifetime (Literal[3600, 10800, 21600, 43200, 86400]): 存活时间
        chart (Chart): 上传谱面

    返回:
        int: 测试服 ID
    '''
    # 转换文件路径并获取名称
    if isinstance(bgm, str):
        bgm = Path(bgm)
    bgm_name = bgm.name
    
    # 构建数据
    data: dict[str, Any] = {
        'title': title,
        'difficulty': str(difficulty),
        'lifetime': str(lifetime),
        'chart': chart.json()
    }
    if hidden:
        data['hidden'] = True
    
    # 准备文件
    file = open(bgm, 'rb')
    files = {'bgm': (bgm_name, file, guess_type(bgm)[0])}
    
    # 发送请求
    response = Api(API['levels'], settings.proxy).request('post', data=data, files=files)
    file.close()
    if (uid := response.json().get('uid', None)) is None:
        raise Exception('上传测试服失败。')
    return uid

# 难度分析
def diff_analysis(
    id_: int,
    diff: Optional[Literal[0, 1, 2, 3, 4]]=None
) -> dict[str, Any]:
    '''难度分析

    参数:
        id_ (int): 谱面 ID
        diff (Optional[Literal[0, 1, 2, 3, 4]], optional): 谱面难度

    返回:
        dict[str, Any]: 难度分析结果
    '''
    # 构建数据
    params = {'id': id_}
    if diff is not None:
        params['diff'] = diff
    
    # 获取结果
    return Api(API['DiffAnalysis'], settings.proxy).request('get', params=params).json()



'''`bestdori.ayachan.utils.utils`

ayachan 工具库模块'''


API = {
    'map-info': 'v2/map-info/',
    'bestdori': 'v2/map-info/bestdori/{id}',
    'DiffAnalysis': 'DiffAnalysis',
    'levels': 'https://sonolus.ayachan.fun/test/sonolus/levels'
}
'''ayachan API 集合'''
